--=======================================================================
function clear_note_value(note_col, probability_value, note_off)
  local probability = tonumber(probability_value)
  print("note_col: " , note_col, "; string: ", note_col.note_string)

  if (note_col.note_string == "---" ) then
    print("No note here so nothing to do")
    return
  end

  local odds = math.random(1,100)

  if odds <= probability then
    print("odds = ", odds, " probability_value = ", probability_value, "; eroding!.")
    if (note_off) then
      note_col.instrument_string = ".."
      note_col.note_string = "OFF" -- "---"
    else
      note_col.instrument_string = ".."
      note_col.note_string = "---"
    end
  else
    print("odds = ", odds, " probability_value = ", probability_value, "; not eroding.")
  end

end




--=======================================================================
-- Takes a list of ints, the number of lines in the pattern, 
-- and the current list of line numbers.
-- The new numbers are created by incrementing from 0
-- and adding each of the `inc_list` values.
--  E.g. 3,2 => 3,5,8,10,13
--  E.g. 3,2,5 => 3,5,10,13,15,20
--=======================================================================
function sequence_from_inc_set(inc_list, lines_in_pattern, current_set, start_line)
  local inc = 0

  print("sequence_from_inc_set is using current set:")
  U.print_pairs(current_set)

  print("sequence_from_inc_set - start_line = " , start_line)

  if start_line  and (start_line < 0) then
    start_line = nil
  end

  if start_line then
    current_set[start_line] = start_line
    inc = start_line
  else
    start_line = 0 
  end

  for i=1+start_line,lines_in_pattern-1 do 
    inc = inc + inc_list[1]

    if (inc >= lines_in_pattern) then
      break
    end

    current_set[inc] = inc
    inc_list = U.wrap(inc_list, 1)

  end

  return current_set
end

--=======================================================================
-- Takes a list of ints, the number of lines in the pattern, 
-- and the current list of line numbers
-- The new numbers are base on whether any LINE_NUM % MOD_INT == 0
--  E.g 3,5 gives 3,5,6,9,10,12,15,18,20, ...
-- Returns that current list with the generated line numbers inserted, 
--    without duplicates
-- That list is unsorted.  
--=======================================================================
function sequence_from_mod_set(mod_list, lines_in_pattern, current_num_set, start_line)

  print("We have been passed current_num_set : ")
  U.print_pairs(current_num_set)

  if tonumber(start_line) < 0 then
    start_line = nil
  end

  if start_line then
    current_num_set[start_line] = start_line
  else
    start_line = 0 -- for the following calcutions
    current_num_set[0] = 0
  end

  for i=1+start_line, lines_in_pattern-1 do 
    for k,v in pairs(mod_list) do
      if not (i <  v) then
        if ( (i%v == 0 )  ) then
          current_num_set[i] = i
        end
      end
    end
  end

  return current_num_set
end


-- ======================================================================
function all_lines()

  local _ti = renoise.song().selected_track_index
  local _fxci = renoise.song().selected_effect_column_index
  local _pi   = renoise.song().selected_pattern_index
  local _col = renoise.song().selected_note_column_index
  local _tp   = renoise.song().patterns[_pi].tracks[_ti]
  local lines_in_pattern = renoise.song().patterns[_pi].number_of_lines
  local all_lines = {}

  for i,l in ipairs(_tp.lines) do
    all_lines[i] = i
  end

  return all_lines
end


-- ======================================================================
function erode(probability_value, lines_list, lines_negation, note_off)

  if #lines_list == 0 then
    lines_list = all_lines()
  end

  local _ti   = renoise.song().selected_track_index
  local _pi   = renoise.song().selected_pattern_index
  local _col  = renoise.song().selected_note_column_index
  local _tp   = renoise.song().patterns[_pi].tracks[_ti]

  local lines_in_pattern = renoise.song().patterns[_pi].number_of_lines

  local odds = 0
  local note_col 

  local adjusted_line_num

  for i,l in ipairs(_tp.lines) do

    adjusted_line_num = i-1
    note_col = l:note_column(_col)

    if (lines_negation) then -- act on line unless in lines list
      if (not U.has_value(lines_list, adjusted_line_num) ) then
        clear_note_value(note_col, probability_value, note_off)
      end
    else -- act on the line if in lines-list
      if ( U.has_value(lines_list, adjusted_line_num) ) then
        clear_note_value(note_col, probability_value, note_off)
      end
    end

  end
  print("lines_negation = ", lines_negation)
end


--=======================================================================
--=======================================================================
Core = {}


-- ======================================================================
function Core.new_set_from_funct_string(function_str, current_set, start_line)
  local _pi   = renoise.song().selected_pattern_index
  local lines_in_pattern = renoise.song().patterns[_pi].number_of_lines
  local func_table = string.to_word_table(function_str)
  local func_char = table.remove(func_table, 1)

  U.str_table_to_int(func_table)
  local generated_line_table = {}
  local sorted_line_numbers = {}

  print("Dispatch on '" .. "'" .. func_char .. "'")

  if (func_char == "+") then
    generated_line_table = sequence_from_inc_set(func_table, lines_in_pattern, current_set, start_line)
  end

  if (func_char == "/") then
    generated_line_table = sequence_from_mod_set(func_table, lines_in_pattern, current_set, start_line)
  end

  for n in pairs(generated_line_table) do table.insert(sorted_line_numbers, n) end

  table.sort(sorted_line_numbers)
  return sorted_line_numbers
end


-- ======================================================================
Core.erode = function(gui)
  local probability_value = string.trim(gui.probability_value_text)
  local lines_list = string.int_list_to_numeric(gui.lines_list)
  erode(probability_value, lines_list, gui.lines_negation, gui.note_off)
end

return Core
