--  Handlers.lua
--  Suggestion: do not put core logic here; try to put that in Core.lua, and
--  just invoke their functions from here.
--  That way those core functions can be used more easily elsewhere,
--  such as by a MIDI-mapping interface. 


-- Some example handlers.  They invoke methods defined in Core.lua
handlers = { 
  {
    pattern = "/set_status_polling",
    handler = function(bool, interval)
      interval = interval or 500 
      print("Handler for /set_status_polling is being passed type ", type(bool) )
--      -- BeatMasher.set_status_polling(bool, interval)
    end 
  }, 

  {
    pattern = "/song/reset",
    handler = function()
      -- BeatMasher.song_reset()
    end 
  }, 

  {
    pattern = "/song/undo",
    handler = function()
      -- BeatMasher.song_undo()
    end 
  },

  {
    pattern = "/track/select", handler = function(track_number)
      -- BeatMasher.track_select(track_number)
    end 
  },
  {
    pattern = "/song/track/clear", handler = function(track_number)
      -- BeatMasher.song_track_clear(track_number)
    end 
  },


  {
    --[[
    To consider: Pass a value for the cloned-track name pre/post-fix.
    Code clones "SomeTrack" to "SomeTrack+"
    If we are cntrolling this postfix ('+') then we can have a function
    to rollback clones (or just delete them and unmute the original.

    --]]
    
    pattern = "/track/clone", handler = function(track_number)
      print("handle track/clone")
      local mute_source_track = true
      -- BeatMasher.clone_track(track_number, mute_source_track)

    end 
  },

  { 
    pattern = "/track/stripe", handler = function(track_number, mod_num)
      print("handle /track/stripe")
      -- BeatMasher.stripe_track(track_number, mod_num)
    end 
  },

  { -- TODO Consider simplyfing names with the assumption that it's always the current selected patten.
    -- e.g. "/track/swap_lines",
    pattern = "/track/swap_lines", handler = function(track_number, mod_num, line_gap)
      print("handle /track/swap_lines")
      -- BeatMasher.swap_lines_pattern_track(track_number, mod_num, line_gap)
    end 
  },

  {
    pattern = "/track/restore", handler = function(track_number)
      print("handle /track/restore")
      -- BeatMasher.restore_track(track_number)
    end 
  },

  {
    pattern = "/trigger/note", handler = function(instrument, track, note,  velocity)
      -- BeatMasher.trigger_note(RENOISE_OSC, instrument, track, note,  velocity)
    end 
  },

  {
    pattern = "/speak/bpm",
    handler = function()
      local instrument_index = 6  -- Would ne nice if this was calculated at runtime
      local track_index = 6    
      -- To keep the core code from having to know too much about creating OSC devices,
      -- pass an existing device into the function.
      --See if this plays out in practice as well as you hope.
      -- BeatMasher.speak_bpm(RENOISE_OSC, track_index, instrument_index)
    end 
  },


} -- end of handlers 

function load_handlers(osc_device)
  for i, h in ipairs(handlers) do
    osc_device:add_message_handler( h.pattern, h.handler )  
  end
--[[
  if (have_rotator) then
    for i, h in ipairs(rotate_handlers) do
      osc_device:add_message_handler( h.pattern, h.handler )  
    end
    print("        ADDED ROTATE HANDLERS")
  else
    print("Cannot add rotate handlers because have_rotator is false")
  end
  --]]

end


-- Experiment to see if tools can add to the default Renoise OSC handlers

local global_action_pattern_map = table.create{}

local instrument_action_pattern_map = table.create{}
local instrument_action_pattern_map_prefix = "/song/instrument"

local track_action_pattern_map = table.create{}
local track_action_pattern_map_prefix = "/song/track"

local device_action_pattern_map = table.create{}
local device_action_pattern_map_prefix = "/device"


local function argument(name, type)
  return { name = name, type = type }
end


  local function add_action(map, info)

    -- validate actions, help finding common errors and typos
    if not (type(info.pattern) == "string" and 
      type(info.handler) == "function") then
      error("An OSC action needs at least a 'pattern' and "..
      "'handler' function property.")
    end

    if not (type(info.description) == "nil" or 
      type(info.description) == "string") then
      error(("OSC action '%s': OSC message description should not be "..
      "specified or should be a string"):format(info.pattern))
    end

    if not (type(info.arguments) == "nil" or 
      type(info.arguments) == "table") then
      error(("OSC action '%s': OSC arguments should not be specified or "..
      "should be a table"):format(info.pattern))
    end

    for _, argument in pairs(info.arguments or {}) do
      if (argument.type ~= "number" and 
        argument.type ~= "string" and
        argument.type ~= "boolean")
        then
          error(("OSC action '%s': unexpected argument type '%s'. "..
          "expected a lua type (number, string or boolean)"):format(
          info.pattern, argument.type or "nil"))
        end
      end

      if (map[info.pattern] ~= nil) then 
        error(("OSC pattern '%s' is already registered"):format(info.pattern))
      end

      -- register the action
      info.arguments = info.arguments or {}
      info.description = info.description or "No description available"

      map[info.pattern] = info
    end


    local function add_global_action(info)
      add_action(global_action_pattern_map, info)
    end


    -- add_instrument_action

    local function add_instrument_action(info)
      add_action(instrument_action_pattern_map, info)
    end


    -- add_track_action

    local function add_track_action(info)
      add_action(track_action_pattern_map, info)
    end


    -- add_device_action

    local function add_device_action(info)
      add_action(device_action_pattern_map, info)
    end

    
    local evaluate_env = {
      _VERSION = _G._VERSION,

      math = table.rcopy(_G.math),
      renoise = table.rcopy(_G.renoise),
      string = table.rcopy(_G.string),
      table = table.rcopy(_G.table),
      assert = _G.assert,
      error = _G.error,
      ipairs = _G.ipairs,
      ripairs = _G.ripairs,
      next = _G.next,
      pairs = _G.pairs,
      pcall = _G.pcall,
      print = _G.print,
      select = _G.select,
      tonumber = _G.tonumber,
      tostring = _G.tostring,
      type = _G.type,
      unpack = _G.unpack,
      xpcall = _G.xpcall
    }


    local song = renoise.song


    add_global_action { 
        pattern = "/ng/erode/test", 
        description = "NG test of setting new Renoise OSC handlers from tools",

        arguments = { argument("msg", "string") },
        handler = function(msg)
          print(" * * * /ng/erode/test has " .. msg) 
        end,
      }

